# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the implementation of AEA agents manager."""
import asyncio
import copy
import os
import threading
from asyncio.tasks import FIRST_COMPLETED
from pathlib import Path
from shutil import rmtree
from threading import Thread
from typing import Callable, Dict, List, Optional

from aea.aea import AEA
from aea.aea_builder import AEABuilder
from aea.configurations.base import AgentConfig, ComponentId, PackageType, PublicId
from aea.configurations.constants import DEFAULT_LEDGER
from aea.configurations.loader import ConfigLoaders
from aea.configurations.project import AgentAlias, Project
from aea.crypto.helpers import create_private_key


class AgentRunAsyncTask:
    """Async task wrapper for agent."""

    def __init__(self, agent: AEA, loop: asyncio.AbstractEventLoop) -> None:
        """Init task with agent and loop."""
        self.run_loop: asyncio.AbstractEventLoop = loop
        self.caller_loop: asyncio.AbstractEventLoop = loop
        self._done_future: Optional[asyncio.Future] = None
        self.task: Optional[asyncio.Task] = None
        self.agent = agent

    def create_run_loop(self) -> None:
        """Create run loop."""
        pass

    def start(self) -> None:
        """Start task."""
        self.create_run_loop()
        self.task = self.run_loop.create_task(self._run_wrapper())
        self._done_future = asyncio.Future(loop=self.caller_loop)

    def wait(self) -> asyncio.Future:
        """Return future to wait task completed."""
        if not self._done_future:  # pragma: nocover
            raise ValueError("Task not started!")
        return self._done_future

    def stop(self) -> None:
        """Stop task."""
        if not self.run_loop or not self.task:  # pragma: nocover
            raise ValueError("Task was not started!")
        self.run_loop.call_soon_threadsafe(self.task.cancel)

    async def _run_wrapper(self) -> None:
        """Run task internals."""
        if not self._done_future:  # pragma: nocover
            raise ValueError("Task was not started! please use start method")
        exc = None
        try:
            await self.run()
        except asyncio.CancelledError:  # pragma: nocover
            pass
        except Exception as e:  # pylint: disable=broad-except
            exc = e
        finally:
            self.caller_loop.call_soon_threadsafe(self._set_result, exc)

    def _set_result(self, exc: Optional[BaseException]) -> None:
        """Set result of task execution."""
        if not self._done_future:  # pragma: nocover
            return
        if exc:
            self._done_future.set_exception(exc)
        else:
            self._done_future.set_result(None)

    async def run(self) -> None:
        """Run task body."""
        self.agent.runtime.set_loop(self.run_loop)
        await self.agent.runtime.run()

    @property
    def is_running(self) -> bool:
        """Return is task running."""
        return not self.wait().done()


class AgentRunThreadTask(AgentRunAsyncTask):
    """Threaded wrapper to run agent."""

    def __init__(self, agent: AEA, loop: asyncio.AbstractEventLoop) -> None:
        """Init task with agent and loop."""
        AgentRunAsyncTask.__init__(self, agent, loop)
        self._thread: Optional[Thread] = None

    def create_run_loop(self) -> None:
        """Create run loop."""
        self.run_loop = asyncio.new_event_loop()

    def start(self) -> None:
        """Run task in a dedicated thread."""
        super().start()
        self._thread = threading.Thread(
            target=self.run_loop.run_until_complete, args=[self.task], daemon=True
        )
        self._thread.start()


class MultiAgentManager:
    """Multi agents manager."""

    AGENT_DO_NOT_OVERRIDE_VALUES = ["skills", "connections", "protocols", "contracts"]
    MODES = ["async", "threaded"]
    DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS = 60

    def __init__(self, working_dir: str, mode: str = "async") -> None:
        """
        Initialize manager.

        :param working_dir: directory to store base agents.
        """
        self.working_dir = working_dir
        self._was_working_dir_created = False
        self._is_running = False
        self._projects: Dict[PublicId, Project] = {}
        self._keys_dir = os.path.abspath(os.path.join(self.working_dir, "keys"))
        self._agents: Dict[str, AgentAlias] = {}
        self._agents_tasks: Dict[str, AgentRunAsyncTask] = {}

        self._thread: Optional[Thread] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._event: Optional[asyncio.Event] = None

        self._error_callbacks: List[Callable[[str, BaseException], None]] = []

        if mode not in self.MODES:
            raise ValueError(
                f'Invalid mode {mode}. Valid modes are {", ".join(self.MODES)}'
            )
        self._started_event = threading.Event()
        self._mode = mode

    @property
    def is_running(self) -> bool:
        """Is manager running."""
        return self._is_running

    def _run_thread(self) -> None:
        """Run internal thread with own event loop."""
        self._loop = asyncio.new_event_loop()
        self._event = asyncio.Event(loop=self._loop)
        self._loop.run_until_complete(self._manager_loop())

    async def _manager_loop(self) -> None:
        """Await and control running manager."""
        if not self._event:
            raise ValueError("Do not use this method directly, use start_manager.")

        self._started_event.set()

        while self._is_running:
            agents_run_tasks_futures = {
                task.wait(): agent_name
                for agent_name, task in self._agents_tasks.items()
            }
            wait_tasks = list(agents_run_tasks_futures.keys()) + [self._event.wait()]  # type: ignore
            done, _ = await asyncio.wait(wait_tasks, return_when=FIRST_COMPLETED)

            if self._event.is_set():
                self._event.clear()

            for task in done:
                if task not in agents_run_tasks_futures:
                    # task not in agents_run_tasks_futures, so it's event_wait, skip it
                    await task
                    continue

                agent_name = agents_run_tasks_futures[task]
                self._agents_tasks.pop(agent_name)
                if task.exception():
                    for callback in self._error_callbacks:
                        callback(agent_name, task.exception())
                else:
                    await task

    def add_error_callback(
        self, error_callback: Callable[[str, BaseException], None]
    ) -> None:
        """Add error callback to call on error raised."""
        self._error_callbacks.append(error_callback)

    def start_manager(self) -> "MultiAgentManager":
        """Start manager."""
        if self._is_running:
            return self

        self._ensure_working_dir()
        self._started_event.clear()
        self._is_running = True
        self._thread = Thread(target=self._run_thread, daemon=True)
        self._thread.start()
        self._started_event.wait(self.DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS)
        return self

    def stop_manager(self) -> "MultiAgentManager":
        """
        Stop manager.

        Stops all running agents and stop agent.

        :return: None
        """
        if not self._is_running:
            return self

        if not self._loop or not self._event or not self._thread:  # pragma: nocover
            raise ValueError("Manager was not started!")

        if not self._thread.is_alive():  # pragma: nocover
            return self

        self.stop_all_agents()

        for agent_name in self.list_agents():
            self.remove_agent(agent_name)

        for project in list(self._projects.keys()):
            self.remove_project(project)

        self._cleanup()

        self._is_running = False

        self._loop.call_soon_threadsafe(self._event.set)

        if self._thread.ident != threading.get_ident():
            self._thread.join(self.DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS)

        self._thread = None
        return self

    def _cleanup(self) -> None:
        """Remove workdir if was created."""
        if self._was_working_dir_created and os.path.exists(self.working_dir):
            rmtree(self.working_dir)

    def add_project(self, public_id: PublicId) -> "MultiAgentManager":
        """Fetch agent project and all dependencies to working_dir."""
        if public_id in self._projects:
            raise ValueError(f"Project {public_id} was already added!")
        self._projects[public_id] = Project.load(self.working_dir, public_id)
        return self

    def remove_project(self, public_id: PublicId) -> "MultiAgentManager":
        """Remove agent project."""
        if public_id not in self._projects:
            raise ValueError(f"Project {public_id} is not present!")

        if self._projects[public_id].agents:
            raise ValueError(
                f"Can not remove projects with aliases exists: {self._projects[public_id].agents}"
            )

        self._projects.pop(public_id).remove()
        return self

    def list_projects(self) -> List[PublicId]:
        """
        List all agents projects added.

        :return: lit of public ids of projects
        """
        return list(self._projects.keys())

    def add_agent(
        self,
        public_id: PublicId,
        agent_name: Optional[str] = None,
        agent_overrides: Optional[dict] = None,
        component_overrides: Optional[List[dict]] = None,
    ) -> "MultiAgentManager":
        """
        Create new agent configuration based on project with config overrides applied.

        Alias is stored in memory only!

        :param public_id: base agent project public id
        :param agent_name: unique name for the agent
        :param agent_overrides: overrides for agent config.
        :param component_overrides: overrides for component section.

        :return: manager
        """
        agent_name = agent_name or public_id.name

        if agent_name in self._agents:
            raise ValueError(f"Agent with name {agent_name} already exists!")

        if public_id not in self._projects:
            raise ValueError(f"{public_id} project is not added!")

        project = self._projects[public_id]

        agent_alias = self._build_agent_alias(
            project=project,
            agent_name=agent_name,
            agent_overrides=agent_overrides,
            component_overrides=component_overrides,
        )
        self._agents[agent_name] = agent_alias
        return self

    def list_agents(self, running_only: bool = False) -> List[str]:
        """
        List all agents.

        :param running_only: returns only running if set to True

        :return: list of agents names
        """
        if running_only:
            return [i for i in self._agents.keys() if self._is_agent_running(i)]
        return list(self._agents.keys())

    def remove_agent(self, agent_name: str) -> "MultiAgentManager":
        """
        Remove agent alias definition from registry.

        :param agent_name: agent name to remove

        :return: None
        """
        if agent_name not in self._agents:
            raise ValueError(f"Agent with name {agent_name} does not exist!")

        if self._is_agent_running(agent_name):
            raise ValueError("Agent is running. stop it first!")

        agent_alias = self._agents.pop(agent_name)
        agent_alias.remove_from_project()
        return self

    def start_agent(self, agent_name: str) -> "MultiAgentManager":
        """
        Start selected agent.

        :param agent_name: agent name to start

        :return: None
        """
        if not self._loop or not self._event:  # pragma: nocover
            raise ValueError("agent is not started!")

        agent_alias = self._agents.get(agent_name)

        if not agent_alias:
            raise ValueError(f"{agent_name} is not registered!")

        if self._is_agent_running(agent_name):
            raise ValueError(f"{agent_name} is already started!")

        if self._mode == "async":
            task = AgentRunAsyncTask(agent_alias.agent, self._loop)
        elif self._mode == "threaded":
            task = AgentRunThreadTask(agent_alias.agent, self._loop)

        task.start()
        self._agents_tasks[agent_name] = task
        self._loop.call_soon_threadsafe(self._event.set)
        return self

    def _is_agent_running(self, agent_name: str) -> bool:
        """Return is agent running state."""
        if agent_name not in self._agents_tasks:
            return False

        task = self._agents_tasks[agent_name]
        return task.is_running

    def start_all_agents(self) -> "MultiAgentManager":
        """
        Start all not started agents.

        :return: None
        """
        self.start_agents(
            [
                agent_name
                for agent_name in self.list_agents()
                if not self._is_agent_running(agent_name)
            ]
        )

        return self

    def stop_agent(self, agent_name: str) -> "MultiAgentManager":
        """
        Stop running agent.

        :param agent_name: agent name to stop

        :return: None
        """
        if not self._is_agent_running(agent_name) or not self._thread or not self._loop:
            raise ValueError(f"{agent_name} is not running!")

        agent_task = self._agents_tasks[agent_name]

        if self._thread.ident == threading.get_ident():  # pragma: nocover
            # In same thread do not perform blocking operations!
            agent_task.stop()
            return self

        wait_future = agent_task.wait()

        event = threading.Event()

        def event_set(*args):  # pylint: disable=unused-argument
            event.set()

        def _add_cb():
            if wait_future.done():
                event_set()  # pragma: nocover
            else:
                wait_future.add_done_callback(event_set)  # pramga: nocover

        self._loop.call_soon_threadsafe(_add_cb)
        agent_task.stop()
        event.wait(self.DEFAULT_TIMEOUT_FOR_BLOCKING_OPERATIONS)

        return self

    def stop_all_agents(self) -> "MultiAgentManager":
        """
        Stop all agents running.

        :return: None
        """
        agents_list = self.list_agents(running_only=True)
        self.stop_agents(agents_list)

        return self

    def stop_agents(self, agent_names: List[str]) -> "MultiAgentManager":
        """
        Stop specified agents.

        :return: None
        """
        for agent_name in agent_names:
            if not self._is_agent_running(agent_name):
                raise ValueError(f"{agent_name} is not running!")

        for agent_name in agent_names:
            self.stop_agent(agent_name)

        return self

    def start_agents(self, agent_names: List[str]) -> "MultiAgentManager":
        """
        Stop specified agents.

        :return: None
        """
        for agent_name in agent_names:
            self.start_agent(agent_name)

        return self

    def get_agent_alias(self, agent_name: str) -> AgentAlias:
        """
        Return details about agent alias definition.

        :return: AgentAlias
        """
        if agent_name not in self._agents:  # pragma: nocover
            raise ValueError(f"Agent with name {agent_name} does not exist!")
        return self._agents[agent_name]

    def _ensure_working_dir(self) -> None:
        """Create working dir if needed."""
        if not os.path.exists(self.working_dir):
            os.makedirs(self.working_dir)
            self._was_working_dir_created = True

        if not os.path.isdir(self.working_dir):  # pragma: nocover
            raise ValueError(f"{self.working_dir} is not a directory!")
        os.makedirs(self._keys_dir)

    def _build_agent_alias(
        self,
        project: Project,
        agent_name: str,
        agent_overrides: Optional[dict] = None,
        component_overrides: Optional[List[dict]] = None,
    ) -> AgentAlias:
        """Create agent alias for project, with given name and overrided values."""
        json_config = self._make_config(
            project.path, agent_overrides, component_overrides
        )

        builder = AEABuilder.from_config_json(json_config, project.path)
        builder.set_name(agent_name)
        builder.set_runtime_mode("threaded")

        if not builder.private_key_paths:
            default_ledger = json_config[0].get("default_ledger", DEFAULT_LEDGER)
            builder.add_private_key(
                default_ledger, self._create_private_key(agent_name, default_ledger)
            )
        agent = builder.build()
        return AgentAlias(project, agent_name, json_config, agent, builder)

    def install_pypi_dependencies(self) -> None:
        """Install dependencies for every project has at least one agent alias."""
        for project in self._projects.values():
            self._install_pypi_dependencies_for_project(project)

    def _install_pypi_dependencies_for_project(self, project: Project) -> None:
        """Install dependencies for project specified if has at least one agent alias."""
        if not project.agents:
            return
        self._install_pypi_dependencies_for_agent(list(project.agents)[0])

    def _install_pypi_dependencies_for_agent(self, agent_name: str) -> None:
        """Install dependencies for the agent registered."""
        self._agents[agent_name].builder.install_pypi_dependencies()

    def _make_config(
        self,
        project_path: str,
        agent_overrides: Optional[dict] = None,
        component_overrides: Optional[List[dict]] = None,
    ) -> List[dict]:
        """Make new config based on project's config with overrides applied."""
        agent_overrides = agent_overrides or {}
        component_overrides = component_overrides or []

        if any([key in agent_overrides for key in self.AGENT_DO_NOT_OVERRIDE_VALUES]):
            raise ValueError(
                'Do not override any of {" ".join(self.AGENT_DO_NOT_OVERRIDE_VALUES)}'
            )

        agent_configuration_file_path: Path = AEABuilder.get_configuration_file_path(
            project_path
        )
        loader = ConfigLoaders.from_package_type(PackageType.AGENT)
        with agent_configuration_file_path.open() as fp:
            agent_config: AgentConfig = loader.load(fp)

        # prepare configuration overrides
        # - agent part
        agent_update_dictionary: Dict = dict(**agent_overrides)
        # - components part
        components_configs: Dict[ComponentId, Dict] = {}
        for obj in component_overrides:
            obj = copy.copy(obj)
            author, name, version = (
                obj.pop("author"),
                obj.pop("name"),
                obj.pop("version"),
            )
            component_id = ComponentId(obj.pop("type"), PublicId(author, name, version))
            components_configs[component_id] = obj
        agent_update_dictionary["component_configurations"] = components_configs
        # do the override (and valiation)
        agent_config.update(agent_update_dictionary)

        # return the multi-paged JSON object.
        json_data = agent_config.ordered_json
        result: List[Dict] = [json_data] + json_data.pop("component_configurations")
        return result

    def _create_private_key(self, name, ledger) -> str:
        """Create new key for agent alias in working dir keys dir."""
        path = os.path.join(self._keys_dir, f"{name}_{ledger}_private.key")
        create_private_key(ledger, path)
        return path



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the scaffold connection."""
import pytest

from aea.configurations.base import ConnectionConfig
from aea.connections.scaffold.connection import MyScaffoldConnection


class TestScaffoldConnectionReception:
    """Test that the stub connection is implemented correctly."""

    def setup(self):
        """Set case up."""
        configuration = ConnectionConfig(
            connection_id=MyScaffoldConnection.connection_id,
        )
        self.connection = MyScaffoldConnection(configuration=configuration)

    @pytest.mark.asyncio
    async def test_methods_not_implemented(self):
        """Test methods not implemented."""
        with pytest.raises(NotImplementedError):
            await self.connection.connect()

        with pytest.raises(NotImplementedError):
            await self.connection.disconnect()

        with pytest.raises(NotImplementedError):
            await self.connection.send(None)

        with pytest.raises(NotImplementedError):
            await self.connection.receive()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""
Updates package versions relative to last release.

Run this script from the root of the project directory:

    python scripts/update_package_versions.py

"""

import argparse
import operator
import os
import pprint
import re
import subprocess  # nosec
import sys
from collections import Counter
from pathlib import Path
from typing import Dict, List, Optional, Pattern, Set

import semver
import yaml
from click.testing import CliRunner

from aea.cli import cli
from aea.configurations.base import PackageId, PackageType, PublicId
from aea.configurations.loader import ConfigLoader
from scripts.generate_ipfs_hashes import update_hashes


DIRECTORIES = ["packages", "aea", "docs", "benchmark", "examples", "tests"]
CLI_LOG_OPTION = ["-v", "OFF"]
TYPES = set(map(lambda x: x.to_plural(), PackageType))
HASHES_CSV = "hashes.csv"
TYPE_TO_CONFIG_FILE = {
    "connections": "connection.yaml",
    "protocols": "protocol.yaml",
    "contracts": "contract.yaml",
    "skills": "skill.yaml",
    "agents": "aea-config.yaml",
}
PUBLIC_ID_REGEX = PublicId.PUBLIC_ID_REGEX[1:-1]
TEST_PROTOCOLS = ["t_protocol", "t_protocol_no_ct"]


def get_protocol_specification_header_regex(public_id: PublicId) -> Pattern:
    """Get the regex to match."""
    return re.compile(
        rf"(name: {public_id.name}\n"
        + rf"author: {public_id.author}\n)"
        + rf"version: {public_id.version}\n"
        + r"(description:)",
        re.MULTILINE,
    )


def check_positive(value):
    """Check value is an int."""
    try:
        ivalue = int(value)
        assert ivalue <= 0
    except (AssertionError, ValueError):
        raise argparse.ArgumentTypeError(f"{value} is an invalid positive int value")
    return ivalue


def parse_arguments():
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-n",
        "--no-interactive",
        action="store_true",
        default=False,
        help="Don't ask user confirmation for replacement.",
    )
    parser.add_argument(
        "-C",
        "--context",
        type=check_positive,
        default=3,
        help="The number of above/below rows to display",
    )

    parser.add_argument(
        "-r",
        "--replace-by-default",
        action="store_true",
        default=False,
        help="If --no-interactive is set, apply the replacement (default: False).",
    )
    return parser.parse_args()


arguments: argparse.Namespace = None  # type: ignore


def check_if_running_allowed() -> None:
    """
    Check if we can run the script.

    Script should only be run on a clean branch.
    """
    git_call = subprocess.Popen(["git", "diff"], stdout=subprocess.PIPE)  # nosec
    (stdout, _) = git_call.communicate()
    git_call.wait()
    if len(stdout) > 0:
        print("Cannot run script in unclean git state.")
        sys.exit(1)


def run_hashing() -> None:
    """Run the hashing script."""
    hashing_call = update_hashes()
    if hashing_call == 1:
        print("Problem when running IPFS script!")
        sys.exit(1)


def get_hashes_from_last_release() -> Dict[str, str]:
    """Get hashes from last release."""
    svn_call = subprocess.Popen(  # nosec
        [
            "svn",
            "export",
            "https://github.com/fetchai/agents-aea.git/trunk/packages/{}".format(
                HASHES_CSV
            ),
        ]
    )
    svn_call.wait()
    hashes = {}  # Dict[str, str]
    with open(HASHES_CSV) as f:
        for line in f:
            split = line.split(",")
            hashes[split[0]] = split[1].rstrip()
    os.remove(HASHES_CSV)
    return hashes


def get_hashes_from_current_release() -> Dict[str, str]:
    """Get hashes from last release."""
    hashes = {}  # Dict[str, str]
    with open(os.path.join("packages", HASHES_CSV)) as f:
        for line in f:
            split = line.split(",")
            hashes[split[0]] = split[1].rstrip()
    return hashes


def split_hashes_by_type(all_hashes: Dict[str, str]) -> Dict[str, Dict[str, str]]:
    """Split hashes by type."""
    result = {
        "agents": {},
        "protocols": {},
        "contracts": {},
        "connections": {},
        "skills": {},
    }  # type: Dict[str, Dict[str, str]]
    for key, value in all_hashes.items():
        if "fetchai" not in key:
            print("Non-fetchai packages not allowed!")
            sys.exit(1)
        _, type_, name = key.split("/")
        result[type_][name] = value
    return result


def get_configuration_file_path(type_: str, name: str) -> Path:
    """Get the configuration file path."""
    fp = os.path.join("packages", "fetchai", type_, name, TYPE_TO_CONFIG_FILE[type_])
    if os.path.isfile(fp):
        return Path(fp)
    fp = os.path.join("aea", type_, name, TYPE_TO_CONFIG_FILE[type_])
    if os.path.isfile(fp):
        return Path(fp)
    print("Cannot find folder for package `{}` of type `{}`".format(name, type_))
    sys.exit(1)


def unified_yaml_load(configuration_file: Path) -> Dict:
    """
    Load YAML file, unified (both single- and multi-paged).

    :param configuration_file: the configuration file path.
    :return: the data.
    """
    package_type = configuration_file.parent.parent.name
    with configuration_file.open() as fp:
        if package_type != "agents":
            return yaml.safe_load(fp)
        # when it is an agent configuration file,
        # we are interested only in the first page of the YAML,
        # because the dependencies are contained only there.
        data = yaml.safe_load_all(fp)
        return list(data)[0]


def get_public_id_from_yaml(configuration_file_path: Path) -> PublicId:
    """
    Get the public id from yaml.

    :param configuration_file_path: the path to the config yaml
    """
    data = unified_yaml_load(configuration_file_path)
    author = data["author"]
    # handle the case when it's a package or agent config file.
    name = data["name"] if "name" in data else data["agent_name"]
    version = data["version"]
    return PublicId(author, name, version)


def public_id_in_registry(type_: str, name: str) -> PublicId:
    """
    Check if a package id is in the registry.

    :param type_: the package type
    :param name: the name of the package
    :return: public id
    """
    runner = CliRunner()
    result = runner.invoke(
        cli, [*CLI_LOG_OPTION, "search", type_, "--query", name], standalone_mode=False,
    )
    reg = r"({}/{}:{})".format("fetchai", name, PublicId.VERSION_REGEX)
    ids = re.findall(reg, result.output,)
    p_ids = []
    highest = PublicId.from_str("fetchai/{}:0.1.0".format(name))
    for id_ in ids:
        p_id = PublicId.from_str(id_[0])
        p_ids.append(p_id)
        if p_id > highest:
            highest = p_id
    return highest


def get_all_package_ids() -> Set[PackageId]:
    """Get all the package ids in the local repository."""
    result: Set[PackageId] = set()
    now = get_hashes_from_current_release()
    now_by_type = split_hashes_by_type(now)
    for type_, name_to_hashes in now_by_type.items():
        for name, _ in name_to_hashes.items():
            if name in TEST_PROTOCOLS:
                continue
            configuration_file_path = get_configuration_file_path(type_, name)
            public_id = get_public_id_from_yaml(configuration_file_path)
            package_id = PackageId(PackageType(type_[:-1]), public_id)
            result.add(package_id)
    return result


def get_public_ids_to_update() -> Set[PackageId]:
    """
    Get all the public ids to be updated.

    In particular, a package DOES NOT NEED a version bump if:
    - the package is a "scaffold" package;
    - the package is no longer present
    - the package hasn't change since the last release;
    - the public ids of the local package and the package in the registry
      are already the same.
    """
    result: Set[PackageId] = set()
    last = get_hashes_from_last_release()
    now = get_hashes_from_current_release()
    last_by_type = split_hashes_by_type(last)
    now_by_type = split_hashes_by_type(now)
    for type_ in TYPES:
        for key, value in last_by_type[type_].items():
            # if the package is a "scaffold" package, skip;
            if key == "scaffold":
                print("Package `{}` of type `{}` is never bumped!".format(key, type_))
                continue
            # if the package is no longer present, skip;
            if key not in now_by_type[type_]:
                print("Package `{}` of type `{}` no longer present!".format(key, type_))
                continue
            # if the package hasn't change since the last release, skip;
            if now_by_type[type_][key] == value:
                print(
                    "Package `{}` of type `{}` has not changed since last release!".format(
                        key, type_
                    )
                )
                continue
            # load public id in the registry if any
            name = key
            configuration_file_path = get_configuration_file_path(type_, name)
            current_public_id = get_public_id_from_yaml(configuration_file_path)
            deployed_public_id = public_id_in_registry(type_, name)
            difference = minor_version_difference(current_public_id, deployed_public_id)
            # check if the public ids of the local package and the package in the registry are already the same.
            package_info = f"Package `{name}` of type `{type_}`"
            public_id_info = f"current id `{current_public_id}` and deployed id `{deployed_public_id}`"
            if difference == 0:
                print(f"{package_info} needs to be bumped!")
                result.add(PackageId(type_[:-1], current_public_id))
            elif difference == 1:
                print(f"{package_info} already at correct version!")
                continue
            else:
                print(f"{package_info} has {public_id_info}. Error!")
                sys.exit(1)
    return result


def _get_ambiguous_public_ids() -> Set[PublicId]:
    """Get the public ids that are the public ids of more than one package id."""
    all_package_ids = get_all_package_ids()
    return set(
        map(
            operator.itemgetter(0),
            filter(
                lambda x: x[0].name != "scaffold" and x[1] > 1,
                Counter(id_.public_id for id_ in all_package_ids).items(),
            ),
        )
    )


def _sort_in_update_order(package_ids: Set[PackageId]) -> List[PackageId]:
    """
    Sort the set of package id in the order of update.

    In particular, they are sorted from the greatest version number to the lowest.

    The reason is to avoid that consecutive package ids (i.e. whose minors difference is 1)
    gets updated in ascending order, resulting in all the updates collapsing to the greatest version.
    For example, consider two package ids with prefix 'author/package' and with versions
    0.1.0 and 0.2.0, respectively. If we bump first the former and then the latter,
    the new replacements associated to the first updated are taken into account in
    the second update.
    """
    return sorted(
        package_ids,
        key=lambda x: (
            semver.VersionInfo.parse(x.public_id.version),
            x.public_id.author,
            x.public_id.name,
            x.package_type.value,
        ),
        reverse=True,
    )


def process_packages(
    all_package_ids_to_update: Set[PackageId], ambiguous_public_ids: Set[PublicId]
) -> None:
    """Process the package versions."""
    print("*" * 100)

    conflicts = {p.public_id for p in all_package_ids_to_update}.intersection(
        ambiguous_public_ids
    )
    print(f"Ambiguous public ids: {pprint.pformat(ambiguous_public_ids)}")
    print(f"Conflicts with public ids to update: {pprint.pformat(conflicts)}",)

    print("*" * 100)
    print("Start processing.")
    sorted_package_ids_list = _sort_in_update_order(all_package_ids_to_update)
    for package_id in sorted_package_ids_list:
        print("#" * 50)
        print(f"Processing {package_id}")
        is_ambiguous = package_id.public_id in ambiguous_public_ids
        process_package(package_id, is_ambiguous)


def minor_version_difference(
    current_public_id: PublicId, deployed_public_id: PublicId
) -> int:
    """Check the minor version difference."""
    diff = semver.compare(current_public_id.version, deployed_public_id.version)
    return diff


def bump_package_version(
    current_public_id: PublicId,
    configuration_file_path: Path,
    type_: str,
    is_ambiguous: bool = False,
) -> None:
    """
    Bump the version references of the package in the repo.

    Includes, bumping the package itself.
    """
    ver = semver.VersionInfo.parse(current_public_id.version)
    new_version = str(ver.bump_minor())
    new_public_id = PublicId(
        current_public_id.author, current_public_id.name, new_version
    )
    for rootdir in DIRECTORIES:
        for path in Path(rootdir).glob("**/*"):
            if path.is_file() and str(path).endswith((".py", ".yaml", ".md")):
                inplace_change(
                    path, current_public_id, new_public_id, type_, is_ambiguous,
                )

    bump_version_in_yaml(configuration_file_path, type_, new_public_id.version)


def _can_disambiguate_from_context(
    line: str, old_string: str, type_: str
) -> Optional[bool]:
    """
    Check whether we can disambiguate the public id given contextual information.

    For example:
    - whether the public id appears in a line of the form 'aea fetch ...' (we know it's an agent)
    - whether the public id appears in a line of the form 'aea add ...' (we know the component type)
    - whether the type appears in the same line where the public id occurs.

    :return: if True/False, the old string can/cannot be replaced. If None, we don't know.
    """
    match = re.search(
        fr"aea +add +(skill|protocol|connection|contract) +{old_string}", line
    )
    if match is not None:
        return match.group(1) == type_[:-1]
    if re.search(fr"aea +fetch +{old_string}", line) is not None:
        return type_ == "agents"
    match = re.search(
        "(skill|SKILL|"
        + "protocol|PROTOCOL|"
        + "connection|CONNECTION|"
        + "contract|CONTRACT|"
        + "agent|AGENT"
        + f")s?.*{old_string}",
        line,
    )
    if match is not None:
        return (match.group(1) + "s") == type_
    return None


def _ask_to_user(lines, line, idx, old_string, type_):
    print("=" * 50)
    above_rows = lines[idx - arguments.context : idx]
    below_rows = lines[idx + 1 : idx + arguments.context]
    print("".join(above_rows))
    print(line.rstrip().replace(old_string, "\033[91m" + old_string + "\033[0m"))
    print("".join(below_rows))
    answer = input(f"Replace for component ({type_}, {old_string})? [y/N]: ",)  # nosec
    return answer


def _ask_to_user_and_replace_if_allowed(content, old_string, new_string, type_) -> str:
    """
    Ask to user if the line should be replaced or not, If the script arguments allow that.

    :param content: the content.
    :param old_string: the old string.
    :param new_string: the new string.
    :param type_: the type of the package.
    :return: the updated content.
    """
    if arguments.no_interactive and arguments.replace_by_default:
        content = content.replace(old_string, new_string)
        return content

    lines = content.splitlines(keepends=True)
    for idx, line in enumerate(lines[:]):
        if old_string not in line:
            continue

        can_replace = _can_disambiguate_from_context(line, old_string, type_)
        # if we managed to replace all the occurrences, then save this line and continue
        if can_replace is not None:
            lines[idx] = line.replace(old_string, new_string) if can_replace else line
            continue

        # otherwise, forget the attempts and ask to the user.
        answer = _ask_to_user(lines, line, idx, old_string, type_)
        if answer == "y":
            lines[idx] = line.replace(old_string, new_string)
    return "".join(lines)


def replace_aea_fetch_statements(content, old_string, new_string, type_) -> str:
    """Replace statements of the type: 'aea fetch <old_string>'."""
    if type_ == "agents":
        content = re.sub(
            fr"aea +fetch +{old_string}", f"aea fetch {new_string}", content
        )
    return content


def replace_aea_add_statements(content, old_string, new_string, type_) -> str:
    """Replace statements of the type: 'aea add <type> <old_string>'."""
    if type_ != "agents":
        content = re.sub(
            fr"aea +add +{type_} +{old_string}",
            f"aea add {type_} {new_string}",
            content,
        )
    return content


def replace_type_and_public_id_occurrences(line, old_string, new_string, type_) -> str:
    """Replace the public id whenever the type and the id occur in the same row, and NOT when other type names occur."""
    if re.match(f"{type_}.*{old_string}", line) and all(
        _type not in line for _type in TYPES.difference({type_})
    ):
        line = line.replace(old_string, new_string)
    return line


def replace_in_yamls(
    content, old_public_id: PublicId, new_public_id: PublicId, type_
) -> str:
    """
    Replace the public id in configuration files (also nested in .md files).

    1) replace package dependencies:
        |protocols:
        |- author/name:version
        |...
        |- old_string
    2) replace in configuration headers:
        |name: package_name
        |author: package_author
        |version: package_version -> bump up
        |type: package_type
    """

    # case 1:
    regex = re.compile(f"({type_}:\n(-.*\n)*)(- *{str(old_public_id)})", re.MULTILINE)
    content = regex.sub(rf"\g<1>- {str(new_public_id)}", content)

    # case 2:
    regex = re.compile(
        rf"(name: {old_public_id.name}\nauthor: {old_public_id.author}\n)version: {old_public_id.version}\n(type: {type_[:-1]})",
        re.MULTILINE,
    )
    content = regex.sub(rf"\g<1>version: {new_public_id.version}\n\g<2>", content)
    return content


def replace_in_protocol_readme(
    fp: Path, content: str, old_public_id: PublicId, new_public_id: PublicId, type_: str
) -> str:
    """
    Replace the version id in the protocol specification in the protcol's README.

    That is, bump the version in cases like:

        |name: package_name
        |author: package_author
        |version: package_version -> bump up
        ...

    :param fp: path to the file being edited.
    :param content: the content of the file.
    :param old_public_id: the old public id.
    :param new_public_id: the new public id.
    :param type_: the type of the package.
    :return: the new content.
    """
    if (
        type_ == fp.parent.parent.name == "protocols"
        and fp.name == "README.md"
        and fp.parent.name == old_public_id.name
    ):
        regex = get_protocol_specification_header_regex(old_public_id)
        content = regex.sub(rf"\g<1>version: {new_public_id.version}\n\g<2>", content)
    return content


def file_should_be_processed(content: str, old_public_id: PublicId):
    """Check if the file should be processed."""
    old_string = str(old_public_id)
    return (
        old_string in content
        or get_protocol_specification_header_regex(old_public_id).search(content)
        is not None
    )


def inplace_change(
    fp: Path,
    old_public_id: PublicId,
    new_public_id: PublicId,
    type_: str,
    is_ambiguous: bool,
) -> None:
    """Replace the occurrence of a string with a new one in the provided file."""

    content = fp.read_text()
    if not file_should_be_processed(content, old_public_id):
        return

    old_string = str(old_public_id)
    new_string = str(new_public_id)
    print(
        f"Processing file {fp} for replacing {old_string} with {new_string} (is_ambiguous: {is_ambiguous})"
    )

    content = replace_in_yamls(content, old_public_id, new_public_id, type_)
    content = replace_in_protocol_readme(
        fp, content, old_public_id, new_public_id, type_
    )
    if not is_ambiguous:
        content = content.replace(old_string, new_string)
    else:
        content = _ask_to_user_and_replace_if_allowed(
            content, old_string, new_string, type_
        )

    with fp.open(mode="w") as f:
        f.write(content)


def bump_version_in_yaml(
    configuration_file_path: Path, type_: str, version: str
) -> None:
    """Bump the package version in the package yaml."""
    loader = ConfigLoader.from_configuration_type(type_[:-1])
    config = loader.load(configuration_file_path.open())
    config.version = version
    loader.dump(config, open(configuration_file_path, "w"))


def process_package(package_id: PackageId, is_ambiguous: bool) -> None:
    """
    Process a package.

    - check version in registry
    - make sure, version is exactly one above the one in registry
    - change all occurrences in packages/tests/aea/examples/benchmark/docs to new reference
    - change yaml version number

    :param package_id: the id of the package
    :param is_ambiguous: whether the public id is ambiguous.
    """
    type_plural = package_id.package_type.to_plural()
    configuration_file_path = get_configuration_file_path(type_plural, package_id.name)
    current_public_id = get_public_id_from_yaml(configuration_file_path)
    bump_package_version(
        current_public_id, configuration_file_path, type_plural, is_ambiguous
    )


def run_once() -> bool:
    """Run the upgrade logic once."""
    all_package_ids_to_update = get_public_ids_to_update()
    if len(all_package_ids_to_update) == 0:
        print("No packages to update. Done!")
        return False
    ambiguous_public_ids = _get_ambiguous_public_ids()
    process_packages(all_package_ids_to_update, ambiguous_public_ids)
    return True


if __name__ == "__main__":
    """
    First, check all hashes are up to date, exit if not.
    Then, run the bumping algo, re-hashing upon each bump.
    """
    arguments = parse_arguments()
    run_hashing()
    check_if_running_allowed()
    while run_once():
        run_hashing()
    sys.exit(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2020 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Serialization module for gym protocol."""

from typing import Any, Dict, cast

from aea.protocols.base import Message, Serializer

from packages.fetchai.protocols.gym import gym_pb2
from packages.fetchai.protocols.gym.custom_types import AnyObject
from packages.fetchai.protocols.gym.message import GymMessage


class GymSerializer(Serializer):
    """Serialization for the 'gym' protocol."""

    @staticmethod
    def encode(msg: Message) -> bytes:
        """
        Encode a 'Gym' message into bytes.

        :param msg: the message object.
        :return: the bytes.
        """
        msg = cast(GymMessage, msg)
        gym_msg = gym_pb2.GymMessage()
        gym_msg.message_id = msg.message_id
        dialogue_reference = msg.dialogue_reference
        gym_msg.dialogue_starter_reference = dialogue_reference[0]
        gym_msg.dialogue_responder_reference = dialogue_reference[1]
        gym_msg.target = msg.target

        performative_id = msg.performative
        if performative_id == GymMessage.Performative.ACT:
            performative = gym_pb2.GymMessage.Act_Performative()  # type: ignore
            action = msg.action
            AnyObject.encode(performative.action, action)
            step_id = msg.step_id
            performative.step_id = step_id
            gym_msg.act.CopyFrom(performative)
        elif performative_id == GymMessage.Performative.PERCEPT:
            performative = gym_pb2.GymMessage.Percept_Performative()  # type: ignore
            step_id = msg.step_id
            performative.step_id = step_id
            observation = msg.observation
            AnyObject.encode(performative.observation, observation)
            reward = msg.reward
            performative.reward = reward
            done = msg.done
            performative.done = done
            info = msg.info
            AnyObject.encode(performative.info, info)
            gym_msg.percept.CopyFrom(performative)
        elif performative_id == GymMessage.Performative.STATUS:
            performative = gym_pb2.GymMessage.Status_Performative()  # type: ignore
            content = msg.content
            performative.content.update(content)
            gym_msg.status.CopyFrom(performative)
        elif performative_id == GymMessage.Performative.RESET:
            performative = gym_pb2.GymMessage.Reset_Performative()  # type: ignore
            gym_msg.reset.CopyFrom(performative)
        elif performative_id == GymMessage.Performative.CLOSE:
            performative = gym_pb2.GymMessage.Close_Performative()  # type: ignore
            gym_msg.close.CopyFrom(performative)
        else:
            raise ValueError("Performative not valid: {}".format(performative_id))

        gym_bytes = gym_msg.SerializeToString()
        return gym_bytes

    @staticmethod
    def decode(obj: bytes) -> Message:
        """
        Decode bytes into a 'Gym' message.

        :param obj: the bytes object.
        :return: the 'Gym' message.
        """
        gym_pb = gym_pb2.GymMessage()
        gym_pb.ParseFromString(obj)
        message_id = gym_pb.message_id
        dialogue_reference = (
            gym_pb.dialogue_starter_reference,
            gym_pb.dialogue_responder_reference,
        )
        target = gym_pb.target

        performative = gym_pb.WhichOneof("performative")
        performative_id = GymMessage.Performative(str(performative))
        performative_content = dict()  # type: Dict[str, Any]
        if performative_id == GymMessage.Performative.ACT:
            pb2_action = gym_pb.act.action
            action = AnyObject.decode(pb2_action)
            performative_content["action"] = action
            step_id = gym_pb.act.step_id
            performative_content["step_id"] = step_id
        elif performative_id == GymMessage.Performative.PERCEPT:
            step_id = gym_pb.percept.step_id
            performative_content["step_id"] = step_id
            pb2_observation = gym_pb.percept.observation
            observation = AnyObject.decode(pb2_observation)
            performative_content["observation"] = observation
            reward = gym_pb.percept.reward
            performative_content["reward"] = reward
            done = gym_pb.percept.done
            performative_content["done"] = done
            pb2_info = gym_pb.percept.info
            info = AnyObject.decode(pb2_info)
            performative_content["info"] = info
        elif performative_id == GymMessage.Performative.STATUS:
            content = gym_pb.status.content
            content_dict = dict(content)
            performative_content["content"] = content_dict
        elif performative_id == GymMessage.Performative.RESET:
            pass
        elif performative_id == GymMessage.Performative.CLOSE:
            pass
        else:
            raise ValueError("Performative not valid: {}.".format(performative_id))

        return GymMessage(
            message_id=message_id,
            dialogue_reference=dialogue_reference,
            target=target,
            performative=performative,
            **performative_content
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Scaffold connection and channel."""
import asyncio
from asyncio import Task
from collections import deque
from typing import Deque, Dict, List, Optional, cast

from aea.connections.base import Connection, ConnectionStates
from aea.mail.base import Envelope
from aea.protocols.base import Message

from packages.fetchai.connections.ledger.base import CONNECTION_ID, RequestDispatcher
from packages.fetchai.connections.ledger.contract_dispatcher import (
    ContractApiRequestDispatcher,
)
from packages.fetchai.connections.ledger.ledger_dispatcher import (
    LedgerApiRequestDispatcher,
)
from packages.fetchai.protocols.contract_api import ContractApiMessage
from packages.fetchai.protocols.ledger_api import LedgerApiMessage


class LedgerConnection(Connection):
    """Proxy to the functionality of the SDK or API."""

    connection_id = CONNECTION_ID

    def __init__(self, **kwargs):
        """Initialize a connection to interact with a ledger APIs."""
        super().__init__(**kwargs)

        self._ledger_dispatcher: Optional[LedgerApiRequestDispatcher] = None
        self._contract_dispatcher: Optional[ContractApiRequestDispatcher] = None
        self._event_new_receiving_task: Optional[asyncio.Event] = None

        self.receiving_tasks: List[asyncio.Future] = []
        self.task_to_request: Dict[asyncio.Future, Envelope] = {}
        self.done_tasks: Deque[asyncio.Future] = deque()
        self.api_configs = self.configuration.config.get(
            "ledger_apis", {}
        )  # type: Dict[str, Dict[str, str]]

    @property
    def event_new_receiving_task(self) -> asyncio.Event:
        """Get the event to notify the 'receive' method of new receiving tasks."""
        return cast(asyncio.Event, self._event_new_receiving_task)

    async def connect(self) -> None:
        """Set up the connection."""

        if self.is_connected:  # pragma: nocover
            return

        self._state.set(ConnectionStates.connecting)

        self._ledger_dispatcher = LedgerApiRequestDispatcher(
            self._state,
            loop=self.loop,
            api_configs=self.api_configs,
            logger=self.logger,
        )
        self._contract_dispatcher = ContractApiRequestDispatcher(
            self._state,
            loop=self.loop,
            api_configs=self.api_configs,
            logger=self.logger,
        )
        self._event_new_receiving_task = asyncio.Event(loop=self.loop)

        self._state.set(ConnectionStates.connected)

    async def disconnect(self) -> None:
        """Tear down the connection."""
        if self.is_disconnected:  # pragma: nocover
            return

        self._state.set(ConnectionStates.disconnecting)

        for task in self.receiving_tasks:
            if not task.cancelled():  # pragma: nocover
                task.cancel()
        self._ledger_dispatcher = None
        self._contract_dispatcher = None
        self._event_new_receiving_task = None

        self._state.set(ConnectionStates.disconnected)

    async def send(self, envelope: "Envelope") -> None:
        """
        Send an envelope.

        :param envelope: the envelope to send.
        :return: None
        """
        task = self._schedule_request(envelope)
        self.receiving_tasks.append(task)
        self.task_to_request[task] = envelope
        self.event_new_receiving_task.set()

    def _schedule_request(self, envelope: Envelope) -> Task:
        """
        Schedule a ledger API request.

        :param envelope: the message.
        :return: None
        """
        dispatcher: RequestDispatcher
        if envelope.protocol_id == LedgerApiMessage.protocol_id:
            if self._ledger_dispatcher is None:  # pragma: nocover
                raise ValueError("No ledger dispatcher set.")
            dispatcher = self._ledger_dispatcher
        elif envelope.protocol_id == ContractApiMessage.protocol_id:
            if self._contract_dispatcher is None:  # pragma: nocover
                raise ValueError("No contract dispatcher set.")
            dispatcher = self._contract_dispatcher
        else:
            raise ValueError("Protocol not supported")

        task = dispatcher.dispatch(envelope)
        return task

    async def receive(self, *args, **kwargs) -> Optional["Envelope"]:
        """
        Receive an envelope. Blocking.

        :return: the envelope received, or None.
        """
        # if there are done tasks, return the result
        if len(self.done_tasks) > 0:  # pragma: nocover
            done_task = self.done_tasks.pop()
            return self._handle_done_task(done_task)

        if len(self.receiving_tasks) == 0:
            self.event_new_receiving_task.clear()
            await self.event_new_receiving_task.wait()

        # wait for completion of at least one receiving task
        done, _ = await asyncio.wait(
            self.receiving_tasks, return_when=asyncio.FIRST_COMPLETED
        )

        # pick one done task
        done_task = done.pop()

        # update done tasks
        self.done_tasks.extend([*done])

        return self._handle_done_task(done_task)

    def _handle_done_task(self, task: asyncio.Future) -> Optional[Envelope]:
        """
        Process a done receiving task.

        :param task: the done task.
        :return: the reponse envelope.
        """
        request = self.task_to_request.pop(task)
        self.receiving_tasks.remove(task)
        response_message: Optional[Message] = task.result()

        response_envelope = None
        if response_message is not None:
            response_envelope = Envelope(
                to=request.sender,
                sender=request.to,
                protocol_id=response_message.protocol_id,
                message=response_message,
                context=request.context,
            )
        return response_envelope



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests of the ethereum module."""

import hashlib
import logging
import time
from unittest.mock import MagicMock

import eth_account
import pytest

from aea.crypto.ethereum import EthereumApi, EthereumCrypto, EthereumFaucetApi

from tests.conftest import (
    ETHEREUM_PRIVATE_KEY_PATH,
    ETHEREUM_TESTNET_CONFIG,
    MAX_FLAKY_RERUNS,
)


def test_creation():
    """Test the creation of the crypto_objects."""
    assert EthereumCrypto(), "Managed to initialise the eth_account"
    assert EthereumCrypto(
        ETHEREUM_PRIVATE_KEY_PATH
    ), "Managed to load the eth private key"


def test_initialization():
    """Test the initialisation of the variables."""
    account = EthereumCrypto()
    assert account.entity is not None, "The property must return the account."
    assert (
        account.address is not None and type(account.address) == str
    ), "After creation the display address must not be None"
    assert (
        account.public_key is not None and type(account.public_key) == str
    ), "After creation the public key must no be None"
    assert account.entity is not None, "After creation the entity must no be None"


def test_derive_address():
    """Test the get_address_from_public_key method"""
    account = EthereumCrypto()
    address = EthereumApi.get_address_from_public_key(account.public_key)
    assert account.address == address, "Address derivation incorrect"


def test_sign_and_recover_message():
    """Test the signing and the recovery function for the eth_crypto."""
    account = EthereumCrypto(ETHEREUM_PRIVATE_KEY_PATH)
    sign_bytes = account.sign_message(message=b"hello")
    assert len(sign_bytes) > 0, "The len(signature) must not be 0"
    recovered_addresses = EthereumApi.recover_message(
        message=b"hello", signature=sign_bytes
    )
    assert len(recovered_addresses) == 1, "Wrong number of addresses recovered."
    assert (
        recovered_addresses[0] == account.address
    ), "Failed to recover the correct address."


def test_sign_and_recover_message_deprecated():
    """Test the signing and the recovery function for the eth_crypto."""
    account = EthereumCrypto(ETHEREUM_PRIVATE_KEY_PATH)
    message = b"hello"
    message_hash = hashlib.sha256(message).digest()
    sign_bytes = account.sign_message(message=message_hash, is_deprecated_mode=True)
    assert len(sign_bytes) > 0, "The len(signature) must not be 0"
    recovered_addresses = EthereumApi.recover_message(
        message=message_hash, signature=sign_bytes, is_deprecated_mode=True
    )
    assert len(recovered_addresses) == 1, "Wrong number of addresses recovered."
    assert (
        recovered_addresses[0] == account.address
    ), "Failed to recover the correct address."


def test_get_hash():
    """Test the get hash functionality."""
    expected_hash = "0x1c8aff950685c2ed4bc3174f3472287b56d9517b9c948127319a09a7a36deac8"
    hash_ = EthereumApi.get_hash(message=b"hello")
    assert expected_hash == hash_


def test_dump_positive():
    """Test dump."""
    account = EthereumCrypto(ETHEREUM_PRIVATE_KEY_PATH)
    account.dump(MagicMock())


def test_api_creation():
    """Test api instantiation."""
    assert EthereumApi(**ETHEREUM_TESTNET_CONFIG), "Failed to initialise the api"


def test_api_none():
    """Test the "api" of the cryptoApi is none."""
    eth_api = EthereumApi(**ETHEREUM_TESTNET_CONFIG)
    assert eth_api.api is not None, "The api property is None."


def test_validate_address():
    """Test the is_valid_address functionality."""
    account = EthereumCrypto()
    assert EthereumApi.is_valid_address(account.address)
    assert not EthereumApi.is_valid_address(account.address + "wrong")


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_balance():
    """Test the balance is zero for a new account."""
    ethereum_api = EthereumApi(**ETHEREUM_TESTNET_CONFIG)
    ec = EthereumCrypto()
    balance = ethereum_api.get_balance(ec.address)
    assert balance == 0, "New account has a positive balance."
    ec = EthereumCrypto(private_key_path=ETHEREUM_PRIVATE_KEY_PATH)
    balance = ethereum_api.get_balance(ec.address)
    assert balance > 0, "Existing account has no balance."


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_construct_sign_and_submit_transfer_transaction():
    """Test the construction, signing and submitting of a transfer transaction."""
    account = EthereumCrypto(private_key_path=ETHEREUM_PRIVATE_KEY_PATH)
    ec2 = EthereumCrypto()
    ethereum_api = EthereumApi(**ETHEREUM_TESTNET_CONFIG)

    amount = 40000
    tx_nonce = ethereum_api.generate_tx_nonce(ec2.address, account.address)
    transfer_transaction = ethereum_api.get_transfer_transaction(
        sender_address=account.address,
        destination_address=ec2.address,
        amount=amount,
        tx_fee=30000,
        tx_nonce=tx_nonce,
        chain_id=3,
    )
    assert (
        isinstance(transfer_transaction, dict) and len(transfer_transaction) == 7
    ), "Incorrect transfer_transaction constructed."

    signed_transaction = account.sign_transaction(transfer_transaction)
    assert (
        isinstance(signed_transaction, eth_account.datastructures.SignedTransaction)
        and len(signed_transaction) == 5
    ), "Incorrect signed_transaction constructed."

    transaction_digest = ethereum_api.send_signed_transaction(signed_transaction)
    assert transaction_digest is not None, "Failed to submit transfer transaction!"

    not_settled = True
    elapsed_time = 0
    while not_settled and elapsed_time < 20:
        elapsed_time += 1
        time.sleep(2)
        transaction_receipt = ethereum_api.get_transaction_receipt(transaction_digest)
        if transaction_receipt is None:
            continue
        is_settled = ethereum_api.is_transaction_settled(transaction_receipt)
        not_settled = not is_settled
    assert transaction_receipt is not None, "Failed to retrieve transaction receipt."
    assert is_settled, "Failed to verify tx!"

    tx = ethereum_api.get_transaction(transaction_digest)
    is_valid = ethereum_api.is_transaction_valid(
        tx, ec2.address, account.address, tx_nonce, amount
    )
    assert is_valid, "Failed to settle tx correctly!"
    assert tx != transaction_receipt, "Should not be same!"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_wealth_positive(caplog):
    """Test the balance is zero for a new account."""
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.ethereum._default_logger"):
        ethereum_faucet_api = EthereumFaucetApi()
        ec = EthereumCrypto()
        ethereum_faucet_api.get_wealth(ec.address)
        assert (
            "Response: " in caplog.text
        ), f"Cannot find message in output: {caplog.text}"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_contract_instance():
    """Test the get contract instance method."""
    contract_address = "0x250A2aeb3eB84782e83365b4c42dbE3CDA9920e4"
    ethereum_api = EthereumApi(**ETHEREUM_TESTNET_CONFIG)
    interface = {"abi": [], "bytecode": b""}
    instance = ethereum_api.get_contract_instance(
        contract_interface=interface, contract_address=contract_address,
    )
    assert str(type(instance)) == "<class 'web3._utils.datatypes.Contract'>"
    instance = ethereum_api.get_contract_instance(contract_interface=interface,)
    assert (
        str(type(instance)) == "<class 'web3._utils.datatypes.PropertyCheckingFactory'>"
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_deploy_transaction():
    """Test the get deploy transaction method."""
    ethereum_api = EthereumApi(**ETHEREUM_TESTNET_CONFIG)
    ec2 = EthereumCrypto()
    interface = {"abi": [], "bytecode": b""}
    deploy_tx = ethereum_api.get_deploy_transaction(
        contract_interface=interface, deployer_address=ec2.address,
    )
    assert type(deploy_tx) == dict and len(deploy_tx) == 6
    assert all(
        key in ["from", "value", "gas", "gasPrice", "nonce", "data"]
        for key in deploy_tx.keys()
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for CLI register command."""

from unittest import TestCase, mock

from aea.cli import cli
from aea.cli.register import do_register
from aea.cli.registry.settings import AUTH_TOKEN_KEY

from tests.conftest import CLI_LOG_OPTION, CliRunner


@mock.patch("aea.cli.register.do_register")
class RegisterTestCase(TestCase):
    """Test case for CLI register command."""

    def setUp(self):
        """Set it up."""
        self.runner = CliRunner()

    def test_register_positive(self, do_register_mock):
        """Test for CLI register positive result."""
        username = "username"
        email = "email@example.com"
        fake_pwd = "fake_pwd"  # nosec

        result = self.runner.invoke(
            cli,
            [
                *CLI_LOG_OPTION,
                "register",
                "--username={}".format(username),
                "--email={}".format(email),
                "--password={}".format(fake_pwd),
                "--confirm_password={}".format(fake_pwd),
            ],
            standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)
        do_register_mock.assert_called_once_with(username, email, fake_pwd, fake_pwd)


@mock.patch("aea.cli.register.validate_author_name", lambda x: x)
@mock.patch("aea.cli.register.register_new_account", return_value="token")
@mock.patch("aea.cli.register.update_cli_config")
class DoRegisterTestCase(TestCase):
    """Test case for do_register method."""

    def test_do_register_positive(self, update_cli_config_mock, *mocks):
        """Test for do_register method positive result."""
        username = "username"
        email = "email@example.com"
        fake_pwd = "fake_pwd"  # nosec

        do_register(username, email, fake_pwd, fake_pwd)
        update_cli_config_mock.assert_called_once_with({AUTH_TOKEN_KEY: "token"})



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test helper to install python dependecies."""
from unittest import mock
from unittest.case import TestCase

from aea.configurations.base import Dependency
from aea.exceptions import AEAException
from aea.helpers.install_dependency import install_dependency


class InstallDependencyTestCase(TestCase):
    """Test case for _install_dependency method."""

    def test__install_dependency_fails(self, *mocks):
        """Test for install_dependency method fails."""
        result = mock.Mock()
        result.returncode = 1
        with mock.patch(
            "aea.helpers.install_dependency.subprocess.Popen", return_value=result
        ):
            with self.assertRaises(AEAException):
                install_dependency("test", Dependency("test", "==10.0.0"))

    def test__install_dependency_ok(self, *mocks):
        """Test for install_dependency method ok."""
        result = mock.Mock()
        result.returncode = 0
        with mock.patch(
            "aea.helpers.install_dependency.subprocess.Popen", return_value=result
        ):
            install_dependency("test", Dependency("test", "==10.0.0"))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the crypto and the ledger APIs registries."""
from typing import Callable, Type

from aea.crypto.base import Crypto, FaucetApi, LedgerApi
from aea.crypto.registries.base import Registry


crypto_registry: Registry[Crypto] = Registry[Crypto]()
register_crypto = crypto_registry.register
make_crypto: Callable[..., Crypto] = crypto_registry.make

ledger_apis_registry: Registry[LedgerApi] = Registry[LedgerApi]()
register_ledger_api = ledger_apis_registry.register
make_ledger_api: Callable[..., LedgerApi] = ledger_apis_registry.make
make_ledger_api_cls: Callable[..., Type[LedgerApi]] = ledger_apis_registry.make_cls

faucet_apis_registry: Registry[FaucetApi] = Registry[FaucetApi]()
register_faucet_api = faucet_apis_registry.register
make_faucet_api: Callable[..., FaucetApi] = faucet_apis_registry.make
make_faucet_api_cls: Callable[..., Type[FaucetApi]] = faucet_apis_registry.make_cls



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for decision_maker."""

import pytest

from aea.decision_maker.scaffold import DecisionMakerHandler
from aea.identity.base import Identity


def test_init_and_not_implemented():
    """Initialise the decision maker handler."""
    decision_maker_handler = DecisionMakerHandler(
        identity=Identity("name", "address"), wallet="wallet"
    )
    with pytest.raises(NotImplementedError):
        decision_maker_handler.handle("message")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the code-blocks in the multiplexer-standalone.md file."""

import logging
import os
from pathlib import Path

from tests.conftest import ROOT_DIR
from tests.test_docs.helper import extract_code_blocks, read_md_file


logger = logging.getLogger(__name__)


class TestDemoDocs:
    """This class contains the tests for the bash/yaml-blocks in *.md file."""

    def test_code_blocks_exist(self):
        """Test that all the code-blocks exist in the python file."""
        path = Path(ROOT_DIR, "tests", "test_docs", "test_bash_yaml", "md_files")
        logger.info(os.listdir(path))
        for file in os.listdir(path):
            if not file.endswith(".md"):
                continue
            bash_file = read_md_file(filepath=Path(path, file))
            md_path = os.path.join(ROOT_DIR, "docs", file.replace("bash-", ""))
            bash_code_blocks = extract_code_blocks(filepath=md_path, filter="bash")
            for blocks in bash_code_blocks:
                assert (
                    blocks in bash_file
                ), "[{}]: FAILED. Code must be identical".format(file)
            logger.info("[{}]: PASSED".format(file))

            yaml_code_blocks = extract_code_blocks(filepath=md_path, filter="yaml")
            for blocks in yaml_code_blocks:
                assert (
                    blocks in bash_file
                ), "[{}]: FAILED. Code must be identical".format(file)
            logger.info("[{}]: PASSED".format(file))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the tests for the `aea gui` sub-commands."""
import json
import shutil
import time
from pathlib import Path
from unittest.mock import patch

from aea.cli.create import create_aea
from aea.cli.utils.context import Context
from aea.test_tools.constants import DEFAULT_AUTHOR

from tests.conftest import CUR_PATH
from tests.test_cli.tools_for_testing import raise_click_exception
from tests.test_cli_gui.test_base import TempCWD, create_app


@patch("aea.cli_gui.cli_create_aea")
def test_create_agent(*mocks):
    """Test creating an agent."""
    app = create_app()
    agent_name = "test_agent_id"

    # Ensure there is now one agent
    response_create = app.post(
        "api/agent", content_type="application/json", data=json.dumps(agent_name)
    )
    assert response_create.status_code == 201
    data = json.loads(response_create.get_data(as_text=True))
    assert data == agent_name


@patch("aea.cli_gui.cli_create_aea", raise_click_exception)
def test_create_agent_fail(*mocks):
    """Test creating an agent and failing."""
    app = create_app()
    agent_name = "test_agent_id"

    response_create = app.post(
        "api/agent", content_type="application/json", data=json.dumps(agent_name)
    )
    assert response_create.status_code == 400
    data = json.loads(response_create.get_data(as_text=True))
    assert data["detail"] == "Failed to create Agent. Message"


def test_real_create_local(*mocks):
    """Really create an agent (have to test the call_aea at some point)."""
    # Set up a temporary current working directory to make agents in
    with TempCWD() as temp_cwd:
        app = create_app()

        # copy the 'packages' directory in the parent of the agent folder.
        shutil.copytree(
            Path(CUR_PATH, "..", "packages"), Path(temp_cwd.temp_dir, "packages")
        )

        agent_id = "test_agent_id"

        # Make an agent
        # We do it programmatically as we need to create an agent with default author
        # that was prevented from GUI.
        ctx = Context(cwd=temp_cwd.temp_dir)
        create_aea(ctx, agent_id, local=True, author=DEFAULT_AUTHOR)

        # Give it a bit of time so the polling funcionts get called
        time.sleep(1)

        # Check that we can actually see this agent too
        response_agents = app.get(
            "api/agent", data=None, content_type="application/json",
        )
        data = json.loads(response_agents.get_data(as_text=True))
        assert response_agents.status_code == 200
        assert len(data) == 1
        assert data[0]["public_id"] == agent_id
        assert data[0]["description"] == "placeholder description"

        # do same but this time find that this is not an agent directory.
        with patch("os.path.isdir", return_value=False):
            response_agents = app.get(
                "api/agent", data=None, content_type="application/json",
            )
        data = json.loads(response_agents.get_data(as_text=True))
        assert response_agents.status_code == 200
        assert len(data) == 0



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for CLI core methods."""

from unittest import TestCase, mock

from click import ClickException

from aea.cli.core import _init_gui
from aea.cli.utils.constants import AUTHOR_KEY


@mock.patch("aea.cli.core.get_or_create_cli_config")
class InitGuiTestCase(TestCase):
    """Test case for _init_gui method."""

    def test__init_gui_positive(self, get_or_create_cli_config_mock):
        """Test _init_gui method for positive result."""
        config = {AUTHOR_KEY: "author"}
        get_or_create_cli_config_mock.return_value = config

        _init_gui()

    def test__init_gui_negative(self, get_or_create_cli_config_mock):
        """Test _init_gui method for negative result."""
        get_or_create_cli_config_mock.return_value = {}
        with self.assertRaises(ClickException):
            _init_gui()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Test module for Registry push methods."""

import os
from unittest import TestCase, mock

from click import ClickException

from aea.cli.registry.push import _compress_dir, _remove_pycache, push_item

from tests.conftest import AUTHOR
from tests.test_cli.tools_for_testing import ContextMock, PublicIdMock


@mock.patch("builtins.open", return_value="opened_file")
@mock.patch("aea.cli.registry.push.check_is_author_logged_in")
@mock.patch("aea.cli.registry.utils._rm_tarfiles")
@mock.patch("aea.cli.registry.push.os.getcwd", return_value="cwd")
@mock.patch("aea.cli.registry.push._compress_dir")
@mock.patch(
    "aea.cli.registry.push.load_yaml",
    return_value={
        "description": "some-description",
        "version": "some-version",
        "author": AUTHOR,
        "protocols": ["protocol_id"],
    },
)
@mock.patch(
    "aea.cli.registry.push.request_api", return_value={"public_id": "public-id"}
)
class PushItemTestCase(TestCase):
    """Test case for push_item method."""

    @mock.patch("aea.cli.registry.push.os.path.exists", return_value=True)
    @mock.patch("aea.cli.registry.push.is_readme_present", return_value=True)
    def test_push_item_positive(
        self,
        is_readme_present_mock,
        path_exists_mock,
        request_api_mock,
        load_yaml_mock,
        compress_mock,
        getcwd_mock,
        rm_tarfiles_mock,
        check_is_author_logged_in_mock,
        open_mock,
    ):
        """Test for push_item positive result."""
        public_id = PublicIdMock(
            name="some-name",
            author="some-author",
            version="{}".format(PublicIdMock.DEFAULT_VERSION),
        )
        push_item(ContextMock(), "some-type", public_id)
        request_api_mock.assert_called_once_with(
            "POST",
            "/some-types/create",
            data={
                "name": "some-name",
                "description": "some-description",
                "version": "some-version",
                "protocols": ["protocol_id"],
            },
            is_auth=True,
            files={"file": "opened_file", "readme": "opened_file"},
        )

    @mock.patch("aea.cli.registry.push.os.path.exists", return_value=True)
    @mock.patch("aea.cli.registry.push.is_readme_present", return_value=False)
    def test_push_item_positive_without_readme(
        self, is_readme_present_mock, path_exists_mock, request_api_mock, *mocks
    ):
        """Test for push_item without readme positive result."""
        public_id = PublicIdMock(
            name="some-name",
            author="some-author",
            version="{}".format(PublicIdMock.DEFAULT_VERSION),
        )
        push_item(ContextMock(), "some-type", public_id)
        request_api_mock.assert_called_once_with(
            "POST",
            "/some-types/create",
            data={
                "name": "some-name",
                "description": "some-description",
                "version": "some-version",
                "protocols": ["protocol_id"],
            },
            is_auth=True,
            files={"file": "opened_file"},
        )

    @mock.patch("aea.cli.registry.push.os.path.exists", return_value=False)
    def test_push_item_item_not_found(
        self,
        path_exists_mock,
        request_api_mock,
        load_yaml_mock,
        compress_mock,
        getcwd_mock,
        rm_tarfiles_mock,
        check_is_author_logged_in_mock,
        open_mock,
    ):
        """Test for push_item - item not found."""
        with self.assertRaises(ClickException):
            push_item(ContextMock(), "some-type", PublicIdMock())

        request_api_mock.assert_not_called()


@mock.patch("aea.cli.registry.push.shutil.rmtree")
class RemovePycacheTestCase(TestCase):
    """Test case for _remove_pycache method."""

    @mock.patch("aea.cli.registry.push.os.path.exists", return_value=True)
    def test_remove_pycache_positive(self, path_exists_mock, rmtree_mock):
        """Test for _remove_pycache positive result."""
        source_dir = "somedir"
        pycache_path = os.path.join(source_dir, "__pycache__")

        _remove_pycache(source_dir)
        rmtree_mock.assert_called_once_with(pycache_path)

    @mock.patch("aea.cli.registry.push.os.path.exists", return_value=False)
    def test_remove_pycache_no_pycache(self, path_exists_mock, rmtree_mock):
        """Test for _remove_pycache if there's no pycache."""
        source_dir = "somedir"
        _remove_pycache(source_dir)
        rmtree_mock.assert_not_called()


@mock.patch("aea.cli.registry.push.tarfile")
@mock.patch("aea.cli.registry.push._remove_pycache")
class CompressDirTestCase(TestCase):
    """Test case for _compress_dir method."""

    def test__compress_dir_positive(self, _remove_pycache_mock, tarfile_mock):
        """Test for _compress_dir positive result."""
        tar_obj_mock = mock.MagicMock()
        open_mock = mock.MagicMock(return_value=tar_obj_mock)
        tarfile_mock.open = open_mock

        _compress_dir("output_filename", "source_dir")
        _remove_pycache_mock.assert_called_once_with("source_dir")
        open_mock.assert_called_once_with("output_filename", "w:gz")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a class representing the game."""

from packages.fetchai.skills.tac_control.game import AgentState as BaseAgentState
from packages.fetchai.skills.tac_control.game import Configuration as BaseConfiguration
from packages.fetchai.skills.tac_control.game import Game as BaseGame
from packages.fetchai.skills.tac_control.game import (
    Initialization as BaseInitialization,
)
from packages.fetchai.skills.tac_control.game import Phase as BasePhase
from packages.fetchai.skills.tac_control.game import Registration as BaseRegistration
from packages.fetchai.skills.tac_control.game import Transaction as BaseTransaction
from packages.fetchai.skills.tac_control.game import Transactions as BaseTransactions


AgentState = BaseAgentState


Configuration = BaseConfiguration


Game = BaseGame


Initialization = BaseInitialization


Phase = BasePhase


Registration = BaseRegistration


Transaction = BaseTransaction


Transactions = BaseTransactions



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the preference representations helper module."""

from aea.helpers.preference_representations.base import (
    linear_utility,
    logarithmic_utility,
)


def test_logarithmic_utility():
    """Test logarithmic utlity."""
    assert (
        logarithmic_utility(
            utility_params_by_good_id={"good_1": 0.2, "good_2": 0.8},
            quantities_by_good_id={"good_1": 2, "good_2": 1},
        )
        > 0
    ), "Utility should be positive."


def test_linear_utility():
    """Test logarithmic utlity."""
    assert (
        linear_utility(
            exchange_params_by_currency_id={"cur_1": 0.2, "cur_2": 0.8},
            balance_by_currency_id={"cur_1": 20, "cur_2": 100},
        )
        > 0
    ), "Utility should be positive."
